@@ -13,6 +13,9 @@ __all__ = [
     "Program",
 ]
 import urllib.parse as urlparse
+from io import BytesIO
+from typing import List
+
 import requests
 import json
 import pprint
@@ -87,10 +90,10 @@ class YesWeHack(YwhApiObject):
 
         return json_resp
 
-    def raw_call(self, method, url, data=None, headers=None):
+    def raw_call(self, method, url, data=None, headers=None, files=None):
         logger.debug("Raw Calling " + url)
         resp = self.session.request(
-            method, url, data=data, headers=headers
+            method, url, data=data, headers=headers, files=files,
         )
         return resp
 
@@ -203,9 +206,15 @@ class YesWeHack(YwhApiObject):
         path = "/reports/{id}/comments".format(id=report_id)
         data = {"private": private, "message": comment, "attachments": []}
         resp = self.call("POST", path, data=json.dumps(data))
+        try:
+            resp.raise_for_status()
+            json_resp = resp.json()
+        except Exception as e:
+            raise InvalidResponse from e
+        return Log(ywh_api=self, **json_resp)
 
     def managed_programs(self, lazy=False):
-        if not managed_pgms or not lazy:
+        if not self.managed_pgms or not lazy:
             self.managed_pgms = [slug['slug'] for slug in self.call('GET', '/manager/programs')['items']]
         return self.managed_pgms
 
@@ -235,17 +244,22 @@ class Attachment(YwhApiObject):
         Attr(name="mime_type", type_=str),
         Attr(name="size", type_=int),
         Attr(name="url", type_=str),
-        Attr(name="data", type_=bytes),
+        Attr(name="_data", type_=bytes, default=bytes()),
     ]
 
     def __init__(self, ywh_api, id=None, lazy=False, **data):
         super().__init__(ywh_api=ywh_api, id=id, **data)
-        if not lazy:
-            self.get_data()
 
-    def get_data(self):
-        r = self.ywh_api.raw_call("GET", self.url)
-        self.data = r.content
+    def load_data(self, force_no_id=False):
+        if not self._data and (force_no_id or self.id is not None):
+            r = self.ywh_api.raw_call("GET", self.url)
+            if r.status_code == 200:
+                self._data = r.content
+
+    @property
+    def data(self) -> bytes:
+        self.load_data()
+        return self._data
 
 
 class BugType(YwhApiObject):
@@ -327,6 +341,7 @@ class Log(YwhApiObject):
         Attr(name="tracker_name", type_=str, default=None),
         Attr(name="tracker_url", type_=str, default=None),
         Attr(name="tracker_id", type_=str, default=None),
+        Attr(name="tracker_token", type_=str, default=None),
     ]
 
     def __init__(self, ywh_api, lazy=False, **kwargs):
@@ -335,8 +350,8 @@ class Log(YwhApiObject):
                 ywh_api=ywh_api, lazy=lazy, **kwargs["author"]
             )
         kwargs["priority"] = Priority(**kwargs["priority"]) if "priority" in kwargs and kwargs['priority'] else None
-        kwargs["old_cvss"] = CVSS(**kwargs["old_cvss"]) if "old_cvss" in kwargs else CVSS()
-        kwargs["new_cvss"] = CVSS(**kwargs["new_cvss"]) if "new_cvss" in kwargs else CVSS()
+        kwargs["old_cvss"] = CVSS(**kwargs["old_cvss"]) if "old_cvss" in kwargs and kwargs["old_cvss"] else CVSS()
+        kwargs["new_cvss"] = CVSS(**kwargs["new_cvss"]) if "new_cvss" in kwargs and kwargs["new_cvss"] else CVSS()
         kwargs["old_bug_type"] = BugType(
             **(kwargs.get("old_bug_type", {}) or {})
         )
@@ -393,6 +408,7 @@ class Report(YwhApiObject):
         Attr(name="source_ips", type_=list, default=[]),
         Attr(name="status", type_=dict, default={}),
         Attr(name="tags", type_=list, default=[]),
+        Attr(name="technical_environment", type_=str),
         Attr(name="technical_information", type_=str),
         Attr(name="technical_information_html", type_=str),
         Attr(name="title", type_=str),
@@ -422,7 +438,7 @@ class Report(YwhApiObject):
         return self.local_id + " : " + self.title
 
     def post_comment(self, comment, private=False):
-        self.ywh_api.post_comment(self.id, comment, private)
+        return self.ywh_api.post_comment(self.id, comment, private)
 
     def get_comments(self, lazy=False):
 
@@ -436,12 +452,12 @@ class Report(YwhApiObject):
 
     def get_attachments_data(self):
         for attachment in self.attachments:
-            attachment.get_data()
+            attachment.load_data(force_no_id=True)
 
     def get_log_attachments_data(self):
         for log in self.logs:
             for attachment in log.attachments:
-                attachment.get_data()
+                attachment.load_data(force_no_id=True)
 
     def get_report_logs(self, lazy=False):
         path = "/reports/{id}/logs".format(id=self.id)
@@ -477,6 +493,109 @@ class Report(YwhApiObject):
         resp = self.ywh_api.raw_call('PUT', f'{self.ywh_api.api_url}/reports/{self.id}/tracking-status', data=json.dumps(datas), headers=headers)
         return resp
 
+    def post_tracker_update(self, tracker_name, tracker_url, tracker_id=None, token=None, message=None):
+        data = {
+            "tracker_name": tracker_name,
+            "tracker_id": tracker_id,
+            "tracker_url": tracker_url,
+            "tracker_token": token or '',
+        }
+        if message is not None:
+            data['message'] = message
+        headers = {**self.ywh_api.session.headers, **self.ywh_api.apps_headers}
+        resp = self.ywh_api.raw_call('POST', f'{self.ywh_api.api_url}/reports/{self.id}/tracker-update', data=json.dumps(data), headers=headers)
+        return resp
+
+    def post_tracker_message(self, tracker_name, tracker_url, tracker_id=None, message=None, attachments=None) -> Log:
+        data = {
+            "tracker_name": tracker_name,
+            "tracker_id": tracker_id,
+            "tracker_url": tracker_url,
+            "attachments": attachments or [],
+        }
+        if message is not None:
+            data['message'] = message
+        headers = {**self.ywh_api.session.headers, **self.ywh_api.apps_headers}
+        http_response = self.ywh_api.raw_call(
+            method='POST',
+            url=f'{self.ywh_api.api_url}/reports/{self.id}/tracker-message',
+            data=json.dumps(data),
+            headers=headers,
+        )
+        if not http_response.ok:
+            raise APIError(f'Unable to send tracker message. Response: {http_response.text}')
+
+        try:
+            content = http_response.json()
+        except ValueError as e:
+            raise APIError(
+                f'Unable to parse response when sending tracker message. Response: {http_response.text}'
+            ) from e
+        return Log(ywh_api=self.ywh_api, **content)
+
+    def post_attachment(self, filename: str, file_content: bytes, file_type: str = None, lazy: bool = False) -> Attachment:
+        headers = {**self.ywh_api.session.headers, **self.ywh_api.apps_headers}
+        http_response = self.ywh_api.raw_call(
+            method='POST',
+            url=f'{self.ywh_api.api_url}/reports/{self.id}/attachments',
+            files={
+                'file': (filename, BytesIO(file_content), file_type)
+            },
+            headers=headers,
+        )
+        if not http_response.ok:
+            raise APIError(f'Unable to create attachment. Response: {http_response.text}')
+
+        try:
+            content = http_response.json()
+        except ValueError as e:
+            raise APIError(
+                f'Unable to parse response when creating attachment. Response: {http_response.text}'
+            ) from e
+        return Attachment(
+            ywh_api=self.ywh_api,
+            id=content['id'],
+            lazy=lazy,
+            **content['attachment'],
+        )
+
+    def put_status(self, status, message=None, attachments=None) -> List[Log]:
+        data = {
+            "status": status,
+            "attachments": attachments or [],
+        }
+        if message is not None:
+            data['message'] = message
+        headers = {**self.ywh_api.session.headers, **self.ywh_api.apps_headers}
+        http_response = self.ywh_api.raw_call(
+            method='PUT',
+            url=f'{self.ywh_api.api_url}/reports/{self.id}/status',
+            data=json.dumps(data),
+            headers=headers,
+        )
+        if not http_response.ok:
+            raise APIError(f'Unable to send report status. Response: {http_response.text}')
+
+        try:
+            content = http_response.json()
+        except ValueError as e:
+            raise APIError(
+                f'Unable to parse response when sending report status. Response: {http_response.text}'
+            ) from e
+
+        if not isinstance(content, dict) or 'items' not in content or not isinstance(content.get('items'), list):
+            raise APIError(
+                f'Unexpected response format when sending report status. Response: {http_response.text}'
+            )
+
+        logs = []
+        for raw_log in content.get('items'):
+            logs.append(Log(ywh_api=self.ywh_api, **raw_log))
+
+        return logs
+
+
+
 class Program(YwhApiObject):
 
     _attrs = [
